import jax.numpy as jnp
import numpy as np


class BasisSpline(object):
    def __init__(
        self,
        n_df,
        knots=None,
        interior_knots=None,
        xrange=(0, 1),
        k=4,
        proper=True,
        normalize=True,
    ):
        self.order = k
        self.N = n_df
        self.xrange = xrange
        if knots is None:
            if interior_knots is None:
                interior_knots = np.linspace(*xrange, n_df - k + 2)
            if proper:
                dx = interior_knots[1] - interior_knots[0]
                knots = np.concatenate(
                    [
                        xrange[0] - dx * np.arange(1, k)[::-1],
                        interior_knots,
                        xrange[1] + dx * np.arange(1, k),
                    ]
                )
            else:
                knots = np.append(
                    np.append(np.array([xrange[0]] * (k - 1)), interior_knots),
                    np.array([xrange[1]] * (k - 1)),
                )
        self.knots = knots
        self.interior_knots = knots
        assert len(self.knots) == self.N + self.order

        self.normalize = normalize
        self.basis_vols = np.ones(self.N)
        if normalize:
            grid = jnp.linspace(*xrange, 1000)
            grid_bases = jnp.array(self.bases(grid))
            self.basis_vols = jnp.array([jnp.trapz(grid_bases[i, :], grid) for i in range(self.N)])

    def norm(self, coefs):
        n = 1.0 / jnp.sum(self.basis_vols * coefs.flatten()) if self.normalize else 1.0
        return n

    def _basis(self, xs, i, k):
        if self.knots[i + k] - self.knots[i] < 1e-6:
            return np.zeros_like(xs)
        elif k == 1:
            v = np.zeros_like(xs)
            v[(xs >= self.knots[i]) & (xs < self.knots[i + 1])] = 1 / (self.knots[i + 1] - self.knots[i])
            return v
        else:
            v = (xs - self.knots[i]) * self._basis(xs, i, k - 1) + (self.knots[i + k] - xs) * self._basis(xs, i + 1, k - 1)
            return (v * k) / ((k - 1) * (self.knots[i + k] - self.knots[i]))

    def _bases(self, xs):
        return [self._basis(xs, i, k=self.order) for i in range(self.N)]

    def bases(self, xs):
        return jnp.concatenate(self._bases(xs)).reshape(self.N, *xs.shape)

    def project(self, bases, coefs):
        coefs /= jnp.sum(coefs)
        return jnp.einsum("i...,i->...", bases, coefs) * self.norm(coefs)

    def eval(self, xs, coefs):
        return self.project(self.bases(xs), coefs)

    def __call__(self, xs, coefs):
        return self.eval(xs, coefs)


class BSpline(BasisSpline):
    def __init__(
        self,
        n_df,
        knots=None,
        interior_knots=None,
        xrange=(0, 1),
        k=4,
        proper=True,
        normalize=False,
    ):
        super().__init__(
            n_df=n_df,
            knots=knots,
            interior_knots=interior_knots,
            xrange=xrange,
            k=k,
            proper=proper,
            normalize=normalize,
        )

    def _bases(self, xs):
        return [(self.knots[i + self.order] - self.knots[i]) / self.order * self._basis(xs, i, k=self.order) for i in range(self.N)]

    def project(self, bases, coefs):
        return jnp.einsum("i...,i->...", bases, coefs) * self.norm(coefs)


class LogXBSpline(BSpline):
    def __init__(self, n_df, knots=None, interior_knots=None, xrange=(0.01, 1), normalize=True, **kwargs):
        knots = None if knots is None else np.log(knots)
        interior_knots = None if interior_knots is None else np.log(interior_knots)
        xrange = np.log(xrange)
        super().__init__(n_df, knots=knots, interior_knots=interior_knots, xrange=xrange, **kwargs)

        self.normalize = normalize
        self.basis_vols = np.ones(self.N)
        if normalize:
            self.grid = jnp.linspace(*np.exp(xrange), 1000)
            self.grid_bases = jnp.array(self.bases(self.grid))

    def bases(self, xs):
        return super().bases(jnp.log(xs))


class LogYBSpline(BSpline):
    def __init__(self, n_df, knots=None, interior_knots=None, xrange=(0, 1), normalize=True, **kwargs):
        super().__init__(n_df, knots=knots, interior_knots=interior_knots, xrange=xrange, **kwargs)
        self.normalize = normalize
        if normalize:
            self.grid = jnp.linspace(*xrange, 1000)
            self.grid_bases = jnp.array(self.bases(self.grid))

    def _project(self, bases, coefs):
        return jnp.exp(jnp.einsum("i...,i->...", bases, coefs))

    def project(self, bases, coefs):
        return self._project(bases, coefs) * self.norm(coefs)

    def norm(self, coefs):
        n = 1.0 / jnp.trapz(self._project(self.grid_bases, coefs), self.grid) if self.normalize else 1.0
        return n


class LogXLogYBSpline(LogYBSpline):
    def __init__(self, n_df, knots=None, interior_knots=None, xrange=(0.1, 1), normalize=True, **kwargs):
        knots = None if knots is None else np.log(knots)
        interior_knots = None if interior_knots is None else np.log(interior_knots)
        xrange = np.log(xrange)
        super().__init__(n_df, knots=knots, interior_knots=interior_knots, xrange=xrange, **kwargs)

        self.normalize = normalize
        self.basis_vols = np.ones(self.N)
        if normalize:
            self.grid = jnp.linspace(*jnp.exp(xrange), 1500)
            self.grid_bases = jnp.array(self.bases(self.grid))

    def bases(self, xs):
        return super().bases(jnp.log(xs))

    def project(self, bases, coefs):
        return self._project(bases, coefs) * self.norm(coefs)


class RectBivariateBasisSpline(object):
    def __init__(
        self,
        xdf,
        ydf,
        xrange=(0, 1),
        yrange=(0, 1),
        kx=4,
        ky=4,
        xbasis=BSpline,
        ybasis=BSpline,
        normalize=True,
    ):
        self.xdf = xdf
        self.ydf = ydf
        self.x_interpolator = xbasis(xdf, xrange=xrange, k=kx, normalize=False)
        self.y_interpolator = ybasis(ydf, xrange=yrange, k=ky, normalize=False)
        self.normalize = normalize
        self.x_bases = None
        self.y_bases = None
        if self.normalize:
            self.gridx = jnp.linspace(*xrange, 750)
            self.gridy = jnp.linspace(*yrange, 750)
            self.gxx, self.gyy = jnp.meshgrid(self.gridx, self.gridy)
            self.grid_bases = self.bases(self.gxx, self.gyy)

    def norm_2d(self, coefs):
        n = 1.0 / jnp.trapz(jnp.trapz(self._project(self.grid_bases, coefs), self.gridy), self.gridx) if self.normalize else 1.0
        return n

    def reset_bases(self):
        self.x_bases = None
        self.y_bases = None

    def bases(self, xs, ys):
        self.x_bases = self.x_interpolator.bases(xs)
        self.y_bases = self.y_interpolator.bases(ys)
        out = jnp.array([[self.x_bases[i] * self.y_bases[j] for i in range(self.xdf)] for j in range(self.ydf)]).reshape(
            self.xdf, self.ydf, *xs.shape
        )
        self.reset_bases()
        return out

    def _project(self, bases, coefs):
        return jnp.exp(jnp.einsum("ij...,ij->...", bases, coefs))

    def project(self, bases, coefs):
        return self._project(bases, coefs) * self.norm_2d(coefs)
